# The LearningOnline Network with CAPA
# (Publication Handler
#
# $Id: loncoursedata.pm,v 1.40 2002/12/11 21:41:02 minaeibi Exp $
#
# Copyright Michigan State University Board of Trustees
#
# This file is part of the LearningOnline Network with CAPA (LON-CAPA).
#
# LON-CAPA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LON-CAPA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LON-CAPA; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# /home/httpd/html/adm/gpl.txt
#
# http://www.lon-capa.org/
#
###

=pod

=head1 NAME

loncoursedata

=head1 SYNOPSIS

Set of functions that download and process student and course information.

=head1 PACKAGES USED

 Apache::Constants qw(:common :http)
 Apache::lonnet()
 Apache::lonhtmlcommon
 HTML::TokeParser
 GDBM_File

=cut

package Apache::loncoursedata;

use strict;
use Apache::Constants qw(:common :http);
use Apache::lonnet();
use Apache::lonhtmlcommon;
use HTML::TokeParser;
use GDBM_File;

=pod

=head1 DOWNLOAD INFORMATION

This section contains all the functions that get data from other servers 
and/or itself.

=cut

# ----- DOWNLOAD INFORMATION -------------------------------------------

=pod

=item &DownloadClasslist()

Collects lastname, generation, middlename, firstname, PID, and section for each
student from their environment database.  The section data is also download, though
it is in a rough format, and is processed later.  The list of students is built from
collecting a classlist for the course that is to be displayed.  Once the classlist
has been downloaded, its date stamp is recorded.  Unless the datestamp for the
class database is reset or is modified, this data will not be downloaded again.  
Also, there was talk about putting the fullname and section
and perhaps other pieces of data into the classlist file.  This would
reduce the number of different file accesses and reduce the amount of 
processing on this side.

=over 4

Input: $courseID, $lastDownloadTime, $c

$courseID:  The id of the course

$lastDownloadTime:  This is the date stamp for when this information was
last gathered.  If it is set to Not downloaded, it will gather the data
again, though it currently does not remove the old data.

$c: The connection class that can determine if the browser has aborted.  It
is used to short circuit this function so that it does not continue to 
get information when there is no need.

Output: \%classlist

\%classlist: A pointer to a hash containing the following data:

-A list of student name:domain (as keys) (known below as $name)

-A hash pointer for each student containing lastname, generation, firstname,
middlename, and PID : Key is $name.studentInformation

-A hash pointer to each students section data : Key is $name.section

-If there was an error in dump, it will be returned in the hash.  See
the error codes for dump in lonnet.  Also, an error key will be 
generated if an abort occurs.

=back

=cut

sub DownloadClasslist {
    my ($courseID, $lastDownloadTime, $c)=@_;
    my ($courseDomain,$courseNumber)=split(/\_/,$courseID);
    my %classlist;

    my $modifiedTime = &Apache::lonnet::GetFileTimestamp($courseDomain, $courseNumber,
                                                         'classlist.db', 
                                                         $Apache::lonnet::perlvar{'lonUsersDir'});

    # Always download the information if lastDownloadTime is set to
    # Not downloaded, otherwise it is only downloaded if the file
    # has been updated and has a more recent date stamp
    if($lastDownloadTime ne 'Not downloaded' &&
       $lastDownloadTime >= $modifiedTime && $modifiedTime >= 0) {
        # Data is not gathered so return UpToDate as true.  This
        # will be interpreted in ProcessClasslist
        $classlist{'lastDownloadTime'}=time;
        $classlist{'UpToDate'} = 'true';
        return \%classlist;
    }

    %classlist=&Apache::lonnet::dump('classlist',$courseDomain, $courseNumber);
    foreach(keys (%classlist)) {
        if(/^(con_lost|error|no_such_host)/i) {
	    return;
        }
    }

    foreach my $name (keys(%classlist)) {
        if(defined($c) && ($c->aborted())) {
            $classlist{'error'}='aborted';
            return \%classlist;
        }

        my ($studentName,$studentDomain) = split(/\:/,$name);
        # Download student environment data, specifically the full name and id.
        my %studentInformation=&Apache::lonnet::get('environment',
                                                    ['lastname','generation',
                                                     'firstname','middlename',
                                                     'id'],
                                                    $studentDomain,
                                                    $studentName);
        $classlist{$name.':studentInformation'}=\%studentInformation;

        if($c->aborted()) {
            $classlist{'error'}='aborted';
            return \%classlist;
        }

        #Section
        my %section=&Apache::lonnet::dump('roles',$studentDomain,$studentName);
        $classlist{$name.':sections'}=\%section;
    }

    $classlist{'UpToDate'} = 'false';
    $classlist{'lastDownloadTime'}=time;

    return \%classlist;
}

=pod

=item &DownloadCourseInformation()

Dump of all the course information for a single student.  The data can be
pruned by making use of dumps regular expression arguement.  This function
also takes a regular expression which it passes straight through to dump.  
The data is no escaped, because it is done elsewhere.  It also
checks the timestamp of the students course database file and only downloads
if it has been modified since the last download.

=over 4

Input: $namedata, $courseID, $lastDownloadTime, $WhatIWant

$namedata: student name:domain

$courseID:  The id of the course

$lastDownloadTime:  This is the date stamp for when this information was
last gathered.  If it is set to Not downloaded, it will gather the data
again, though it currently does not remove the old data.

$WhatIWant:  Regular expression used to get selected data with dump

Output: \%courseData

\%courseData:  A hash pointer to the raw data from the students course
database.

=back

=cut

sub DownloadCourseInformation {
    my ($namedata,$courseID,$lastDownloadTime,$WhatIWant)=@_;
    my %courseData;
    my ($name,$domain) = split(/\:/,$namedata);

    my $modifiedTime = &Apache::lonnet::GetFileTimestamp($domain, $name,
                                      $courseID.'.db', 
                                      $Apache::lonnet::perlvar{'lonUsersDir'});

    if($lastDownloadTime >= $modifiedTime && $modifiedTime >= 0) {
        # Data is not gathered so return UpToDate as true.  This
        # will be interpreted in ProcessClasslist
        $courseData{$namedata.':lastDownloadTime'}=time;
        $courseData{$namedata.':UpToDate'} = 'true';
        return \%courseData;
    }

    # Download course data
    if(!defined($WhatIWant)) {
        # set the regular expression to everything by setting it to period
        $WhatIWant = '.';
    }
    %courseData=&Apache::lonnet::dump($courseID, $domain, $name, $WhatIWant);
    $courseData{'UpToDate'} = 'false';
    $courseData{'lastDownloadTime'}=time;

    my %newData;
    foreach (keys(%courseData)) {
        # need to have the keys to be prepended with the name:domain of the
        # student to reduce data collision later.
        $newData{$namedata.':'.$_} = $courseData{$_};
    }

    return \%newData;
}

# ----- END DOWNLOAD INFORMATION ---------------------------------------

=pod

=head1 PROCESSING FUNCTIONS

These functions process all the data for all the students.  Also, they
are the functions that access the cache database for writing the majority of
the time.  The downloading and caching were separated to reduce problems 
with stopping downloading then can not tie hash to database later.

=cut

# ----- PROCESSING FUNCTIONS ---------------------------------------

=pod

=item &ProcessTopResourceMap()

Trace through the "big hash" created in rat/lonuserstate.pm::loadmap.  
Basically, this function organizes a subset of the data and stores it in
cached data.  The data stored is the problems, sequences, sequence titles,
parts of problems, and their ordering.  Column width information is also 
partially handled here on a per sequence basis.

=over 4

Input: $cache, $c

$cache:  A pointer to a hash to store the information

$c:  The connection class used to determine if an abort has been sent to the 
browser

Output: A string that contains an error message or "OK" if everything went 
smoothly.

=back

=cut

sub ProcessTopResourceMap {
    my ($cache,$c)=@_;
    my %hash;
    my $fn=$ENV{'request.course.fn'};
    if(-e "$fn.db") {
	my $tieTries=0;
	while($tieTries < 3) {
            if($c->aborted()) {
                return;
            }
	    if(tie(%hash,'GDBM_File',"$fn.db",&GDBM_READER(),0640)) {
		last;
	    }
	    $tieTries++;
	    sleep 1;
	}
	if($tieTries >= 3) {
            return 'Coursemap undefined.';
        }
    } else {
        return 'Can not open Coursemap.';
    }

    my $oldkeys;
    delete $cache->{'OptionResponses'};
    if(defined($cache->{'ResourceKeys'})) {
        $oldkeys = $cache->{'ResourceKeys'};
        foreach (split(':::', $cache->{'ResourceKeys'})) {
            delete $cache->{$_};
        }
        delete $cache->{'ResourceKeys'};
    }

    # Initialize state machine.  Set information pointing to top level map.
    my (@sequences, @currentResource, @finishResource);
    my ($currentSequence, $currentResourceID, $lastResourceID);

    $currentResourceID=$hash{'ids_'.
      &Apache::lonnet::clutter($ENV{'request.course.uri'})};
    push(@currentResource, $currentResourceID);
    $lastResourceID=-1;
    $currentSequence=-1;
    my $topLevelSequenceNumber = $currentSequence;

    my %sequenceRecord;
    my %allkeys;
    while(1) {
        if($c->aborted()) {
            last;
        }
	# HANDLE NEW SEQUENCE!
	#if page || sequence
	if(defined($hash{'map_pc_'.$hash{'src_'.$currentResourceID}}) &&
           !defined($sequenceRecord{$currentResourceID})) {
            $sequenceRecord{$currentResourceID}++;
	    push(@sequences, $currentSequence);
	    push(@currentResource, $currentResourceID);
	    push(@finishResource, $lastResourceID);

	    $currentSequence=$hash{'map_pc_'.$hash{'src_'.$currentResourceID}};

            # Mark sequence as containing problems.  If it doesn't, then
            # it will be removed when processing for this sequence is
            # complete.  This allows the problems in a sequence
            # to be outputed before problems in the subsequences
            if(!defined($cache->{'orderedSequences'})) {
                $cache->{'orderedSequences'}=$currentSequence;
            } else {
                $cache->{'orderedSequences'}.=':'.$currentSequence;
            }
            $allkeys{'orderedSequences'}++;

	    $lastResourceID=$hash{'map_finish_'.
				  $hash{'src_'.$currentResourceID}};
	    $currentResourceID=$hash{'map_start_'.
				     $hash{'src_'.$currentResourceID}};

	    if(!($currentResourceID) || !($lastResourceID)) {
		$currentSequence=pop(@sequences);
		$currentResourceID=pop(@currentResource);
		$lastResourceID=pop(@finishResource);
		if($currentSequence eq $topLevelSequenceNumber) {
		    last;
		}
	    }
            next;
	}

	# Handle gradable resources: exams, problems, etc
	$currentResourceID=~/(\d+)\.(\d+)/;
        my $partA=$1;
        my $partB=$2;
	if($hash{'src_'.$currentResourceID}=~
	   /\.(problem|exam|quiz|assess|survey|form)$/ &&
	   $partA eq $currentSequence && 
           !defined($sequenceRecord{$currentSequence.':'.
                                    $currentResourceID})) {
            $sequenceRecord{$currentSequence.':'.$currentResourceID}++;
	    my $Problem = &Apache::lonnet::symbclean(
			  &Apache::lonnet::declutter($hash{'map_id_'.$partA}).
			  '___'.$partB.'___'.
			  &Apache::lonnet::declutter($hash{'src_'.
							 $currentResourceID}));

	    $cache->{$currentResourceID.':problem'}=$Problem;
            $allkeys{$currentResourceID.':problem'}++;
	    if(!defined($cache->{$currentSequence.':problems'})) {
		$cache->{$currentSequence.':problems'}=$currentResourceID;
	    } else {
		$cache->{$currentSequence.':problems'}.=
		    ':'.$currentResourceID;
	    }
            $allkeys{$currentSequence.':problems'}++;

	    my $meta=$hash{'src_'.$currentResourceID};
#            $cache->{$currentResourceID.':title'}=
#                &Apache::lonnet::metdata($meta,'title');
            $cache->{$currentResourceID.':title'}=
                $hash{'title_'.$currentResourceID};
            $allkeys{$currentResourceID.':title'}++;
            $cache->{$currentResourceID.':source'}=
                $hash{'src_'.$currentResourceID};
            $allkeys{$currentResourceID.':source'}++;

            # Get Parts for problem
            my %beenHere;
            foreach (split(/\,/,&Apache::lonnet::metadata($meta,'packages'))) {
                if(/^\w+response_\d+.*/) {
                    my (undef, $partId, $responseId) = split(/_/,$_);
                    if($beenHere{'p:'.$partId} ==  0) {
                        $beenHere{'p:'.$partId}++;
                        if(!defined($cache->{$currentSequence.':'.
                                            $currentResourceID.':parts'})) {
                            $cache->{$currentSequence.':'.$currentResourceID.
                                     ':parts'}=$partId;
                        } else {
                            $cache->{$currentSequence.':'.$currentResourceID.
                                     ':parts'}.=':'.$partId;
                        }
                        $allkeys{$currentSequence.':'.$currentResourceID.
                                  ':parts'}++;
                    }
                    if($beenHere{'r:'.$partId.':'.$responseId} == 0) {
                        $beenHere{'r:'.$partId.':'.$responseId}++;
                        if(!defined($cache->{$currentSequence.':'.
                                             $currentResourceID.':'.$partId.
                                             ':responseIDs'})) {
                            $cache->{$currentSequence.':'.$currentResourceID.
                                     ':'.$partId.':responseIDs'}=$responseId;
                        } else {
                            $cache->{$currentSequence.':'.$currentResourceID.
                                     ':'.$partId.':responseIDs'}.=':'.
                                                                  $responseId;
                        }
                        $allkeys{$currentSequence.':'.$currentResourceID.':'.
                                     $partId.':responseIDs'}++;
                    }
                    if(/^optionresponse/ && 
                       $beenHere{'o:'.$partId.':'.$currentResourceID} == 0) {
                        $beenHere{'o:'.$partId.$currentResourceID}++;
                        if(defined($cache->{'OptionResponses'})) {
                            $cache->{'OptionResponses'}.= ':::'.
                                $currentSequence.':'.$currentResourceID.':'.
                                $partId.':'.$responseId;
                        } else {
                            $cache->{'OptionResponses'}= $currentSequence.':'.
                                $currentResourceID.':'.
                                $partId.':'.$responseId;
                        }
                        $allkeys{'OptionResponses'}++;
                    }
                }
            }
        }

	# if resource == finish resource, then it is the end of a sequence/page
	if($currentResourceID eq $lastResourceID) {
	    # pop off last resource of sequence
	    $currentResourceID=pop(@currentResource);
	    $lastResourceID=pop(@finishResource);

	    if(defined($cache->{$currentSequence.':problems'})) {
		# Capture sequence information here
		$cache->{$currentSequence.':title'}=
		    $hash{'title_'.$currentResourceID};
                $allkeys{$currentSequence.':title'}++;
                $cache->{$currentSequence.':source'}=
                    $hash{'src_'.$currentResourceID};
                $allkeys{$currentSequence.':source'}++;

                my $totalProblems=0;
                foreach my $currentProblem (split(/\:/,
                                               $cache->{$currentSequence.
                                               ':problems'})) {
                    foreach (split(/\:/,$cache->{$currentSequence.':'.
                                                   $currentProblem.
                                                   ':parts'})) {
                        $totalProblems++;
                    }
                }
		my @titleLength=split(//,$cache->{$currentSequence.
                                                    ':title'});
                # $extra is 5 for problems correct and 3 for space
                # between problems correct and problem output
                my $extra = 8;
		if(($totalProblems + $extra) > (scalar @titleLength)) {
		    $cache->{$currentSequence.':columnWidth'}=
                        $totalProblems + $extra;
		} else {
		    $cache->{$currentSequence.':columnWidth'}=
                        (scalar @titleLength);
		}
                $allkeys{$currentSequence.':columnWidth'}++;
	    } else {
                # Remove sequence from list, if it contains no problems to
                # display.
                $cache->{'orderedSequences'}=~s/$currentSequence//;
                $cache->{'orderedSequences'}=~s/::/:/g;
                $cache->{'orderedSequences'}=~s/^:|:$//g;
            }

	    $currentSequence=pop(@sequences);
	    if($currentSequence eq $topLevelSequenceNumber) {
		last;
	    }
        }

	# MOVE!!!
	# move to next resource
	unless(defined($hash{'to_'.$currentResourceID})) {
	    # big problem, need to handle.  Next is probably wrong
            my $errorMessage = 'Big problem in ';
            $errorMessage .= 'loncoursedata::ProcessTopLevelMap.';
            $errorMessage .= '  bighash to_$currentResourceID not defined!';
            &Apache::lonnet::logthis($errorMessage);
	    last;
	}
	my @nextResources=();
	foreach (split(/\,/,$hash{'to_'.$currentResourceID})) {
            if(!defined($sequenceRecord{$currentSequence.':'.
                                        $hash{'goesto_'.$_}})) {
                push(@nextResources, $hash{'goesto_'.$_});
            }
	}
	push(@currentResource, @nextResources);
	# Set the next resource to be processed
	$currentResourceID=pop(@currentResource);
    }

    my @theKeys = keys(%allkeys);
    my $newkeys = join(':::', @theKeys);
    $cache->{'ResourceKeys'} = join(':::', $newkeys);
    if($newkeys ne $oldkeys) {
        $cache->{'ResourceUpdated'} = 'true';
    } else {
        $cache->{'ResourceUpdated'} = 'false';
    }

    unless (untie(%hash)) {
        &Apache::lonnet::logthis("<font color=blue>WARNING: ".
                                 "Could not untie coursemap $fn (browse)".
                                 ".</font>"); 
    }

    return 'OK';
}

=pod

=item &ProcessClasslist()

Taking the class list dumped from &DownloadClasslist(), all the 
students and their non-class information is processed using the 
&ProcessStudentInformation() function.  A date stamp is also recorded for
when the data was processed.

Takes data downloaded for a student and breaks it up into managable pieces and 
stored in cache data.  The username, domain, class related date, PID, 
full name, and section are all processed here.

=over 4

Input: $cache, $classlist, $courseID, $ChartDB, $c

$cache: A hash pointer to store the data

$classlist:  The hash of data collected about a student from 
&DownloadClasslist().  The hash contains a list of students, a pointer 
to a hash of student information for each student, and each students section 
number.

$courseID:  The course ID

$ChartDB:  The name of the cache database file.

$c:  The connection class used to determine if an abort has been sent to the 
browser

Output: @names

@names:  An array of students whose information has been processed, and are to 
be considered in an arbitrary order.  The entries in @names are of the form
username:domain.

The values in $cache are as follows:

 *NOTE: for the following $name implies username:domain
 $name.':error'                  only defined if an error occured.  Value
                                 contains the error message
 $name.':lastDownloadTime'       unconverted time of the last update of a
                                 student\'s course data
 $name.'updateTime'              coverted time of the last update of a 
                                 student\'s course data
 $name.':username'               username of a student
 $name.':domain'                 domain of a student
 $name.':fullname'               full name of a student
 $name.':id'                     PID of a student
 $name.':Status'                 active/expired status of a student
 $name.':section'                section of a student

=back

=cut

sub ProcessClasslist {
    my ($cache,$classlist,$courseID,$c)=@_;
    my @names=();

    $cache->{'ClasslistTimeStamp'}=$classlist->{'lastDownloadTime'};
    if($classlist->{'UpToDate'} eq 'true') {
        return split(/:::/,$cache->{'NamesOfStudents'});;
    }

    foreach my $name (keys(%$classlist)) {
        if($name =~ /\:section/ || $name =~ /\:studentInformation/ ||
           $name eq '' || $name eq 'UpToDate' || $name eq 'lastDownloadTime') {
            next;
        }
        if($c->aborted()) {
            return ();
        }
        my $studentInformation = $classlist->{$name.':studentInformation'};
        my $date = $classlist->{$name};
        my ($studentName,$studentDomain) = split(/\:/,$name);

        $cache->{$name.':username'}=$studentName;
        $cache->{$name.':domain'}=$studentDomain;
        # Initialize timestamp for student
        if(!defined($cache->{$name.':lastDownloadTime'})) {
            $cache->{$name.':lastDownloadTime'}='Not downloaded';
            $cache->{$name.':updateTime'}=' Not updated';
        }

        my $error = 0;
        foreach(keys(%$studentInformation)) {
            if(/^(con_lost|error|no_such_host)/i) {
                $cache->{$name.':error'}=
                    'Could not download student environment data.';
                $cache->{$name.':fullname'}='';
                $cache->{$name.':id'}='';
                $error = 1;
            }
        }
        next if($error);
        push(@names,$name);
        $cache->{$name.':fullname'}=&ProcessFullName(
                                          $studentInformation->{'lastname'},
                                          $studentInformation->{'generation'},
                                          $studentInformation->{'firstname'},
                                          $studentInformation->{'middlename'});
        $cache->{$name.':id'}=$studentInformation->{'id'};

        my ($end, $start)=split(':',$date);
        $courseID=~s/\_/\//g;
        $courseID=~s/^(\w)/\/$1/;

        my $sec='';
        my $sectionData = $classlist->{$name.':sections'};
        foreach my $key (keys (%$sectionData)) {
            my $value = $sectionData->{$key};
            if ($key=~/^$courseID(?:\/)*(\w+)*\_st$/) {
                my $tempsection=$1;
                if($key eq $courseID.'_st') {
                    $tempsection='';
                }
                my (undef,$roleend,$rolestart)=split(/\_/,$value);
                if($roleend eq $end && $rolestart eq $start) {
                    $sec = $tempsection;
                    last;
                }
            }
        }

        my $status='Expired';
        if(((!$end) || time < $end) && ((!$start) || (time > $start))) {
            $status='Active';
        }
        $cache->{$name.':Status'}=$status;
        $cache->{$name.':section'}=$sec;

        if($sec eq '' || !defined($sec) || $sec eq ' ') {
            $sec = 'none';
        }
        if(defined($cache->{'sectionList'})) {
            if($cache->{'sectionList'} !~ /(^$sec:|^$sec$|:$sec$|:$sec:)/) {
                $cache->{'sectionList'} .= ':'.$sec;
            }
        } else {
            $cache->{'sectionList'} = $sec;
        }
    }

    $cache->{'ClasslistTimestamp'}=time;
    $cache->{'NamesOfStudents'}=join(':::',@names);

    return @names;
}

=pod

=item &ProcessStudentData()

Takes the course data downloaded for a student in 
&DownloadCourseInformation() and breaks it up into key value pairs
to be stored in the cached data.  The keys are comprised of the 
$username:$domain:$keyFromCourseDatabase.  The student username:domain is
stored away signifying that the students information has been downloaded and 
can be reused from cached data.

=over 4

Input: $cache, $courseData, $name

$cache: A hash pointer to store data

$courseData:  A hash pointer that points to the course data downloaded for a 
student.

$name:  username:domain

Output: None

*NOTE:  There is no output, but an error message is stored away in the cache 
data.  This is checked in &FormatStudentData().  The key username:domain:error 
will only exist if an error occured.  The error is an error from 
&DownloadCourseInformation().

=back

=cut

sub ProcessStudentData {
    my ($cache,$courseData,$name)=@_;

    if(!&CheckDateStampError($courseData, $cache, $name)) {
        return;
    }

    # This little delete thing, should not be here.  Move some other
    # time though.
    if(defined($cache->{$name.':keys'})) {
	foreach (split(':::', $cache->{$name.':keys'})) {
	    delete $cache->{$name.':'.$_};
	}
        delete $cache->{$name.':keys'};
    }

    my %courseKeys;
    # user name:domain was prepended earlier in DownloadCourseInformation
    foreach (keys %$courseData) {
	my $currentKey = $_;
	$currentKey =~ s/^$name//;
	$courseKeys{$currentKey}++;
        $cache->{$_}=$courseData->{$_};
    }

    $cache->{$name.':keys'} = join(':::', keys(%courseKeys));

    return;
}

=pod

=item &ExtractStudentData()

HISTORY: This function originally existed in every statistics module,
and performed different tasks, the had some overlap.  Due to the need
for the data from the different modules, they were combined into
a single function.

This function now extracts all the necessary course data for a student
from what was downloaded from their homeserver.  There is some extra
time overhead compared to the ProcessStudentInformation function, but
it would have had to occurred at some point anyways.  This is now
typically called while downloading the data it will process.  It is
the brother function to ProcessStudentInformation.

=over 4

Input: $input, $output, $data, $name

$input: A hash that contains the input data to be processed

$output: A hash to contain the processed data

$data: A hash containing the information on what is to be
processed and how (basically).

$name:  username:domain

The input is slightly different here, but is quite simple.
It is currently used where the $input, $output, and $data
can and are often the same hashes, but they do not need
to be.

Output: None

*NOTE:  There is no output, but an error message is stored away in the cache 
data.  This is checked in &FormatStudentData().  The key username:domain:error 
will only exist if an error occured.  The error is an error from 
&DownloadCourseInformation().

=back

=cut

sub ExtractStudentData {
    my ($input, $output, $data, $name)=@_;

    if(!&CheckDateStampError($input, $data, $name)) {
        return;
    }

    # This little delete thing, should not be here.  Move some other
    # time though.
    my %allkeys;
    if(defined($output->{$name.':keys'})) {
	foreach (split(':::', $output->{$name.':keys'})) {
	    delete $output->{$name.':'.$_};
	}
        delete $output->{$name.':keys'};
    }

    my ($username,$domain)=split(':',$name);

    my $Version;
    my $problemsCorrect = 0;
    my $totalProblems   = 0;
    my $problemsSolved  = 0;
    my $numberOfParts   = 0;
    my $totalAwarded    = 0;
    foreach my $sequence (split(':', $data->{'orderedSequences'})) {
        foreach my $problemID (split(':', $data->{$sequence.':problems'})) {
            my $problem = $data->{$problemID.':problem'};
            my $LatestVersion = $input->{$name.':version:'.$problem};

            # Output dashes for all the parts of this problem if there
            # is no version information about the current problem.
            $output->{$name.':'.$problemID.':NoVersion'} = 'false';
            $allkeys{$name.':'.$problemID.':NoVersion'}++;
            if(!$LatestVersion) {
                foreach my $part (split(/\:/,$data->{$sequence.':'.
                                                      $problemID.
                                                      ':parts'})) {
                    $output->{$name.':'.$problemID.':'.$part.':tries'} = 0;
                    $output->{$name.':'.$problemID.':'.$part.':awarded'} = 0;
                    $output->{$name.':'.$problemID.':'.$part.':code'} = ' ';
		    $allkeys{$name.':'.$problemID.':'.$part.':tries'}++;
		    $allkeys{$name.':'.$problemID.':'.$part.':awarded'}++;
		    $allkeys{$name.':'.$problemID.':'.$part.':code'}++;
                    $totalProblems++;
                }
                $output->{$name.':'.$problemID.':NoVersion'} = 'true';
                next;
            }

            my %partData=undef;
            # Initialize part data, display skips correctly
            # Skip refers to when a student made no submissions on that
            # part/problem.
            foreach my $part (split(/\:/,$data->{$sequence.':'.
                                                 $problemID.
                                                 ':parts'})) {
                $partData{$part.':tries'}=0;
                $partData{$part.':code'}=' ';
                $partData{$part.':awarded'}=0;
                $partData{$part.':timestamp'}=0;
                foreach my $response (split(':', $data->{$sequence.':'.
                                                         $problemID.':'.
                                                         $part.':responseIDs'})) {
                    $partData{$part.':'.$response.':submission'}='';
                }
            }

            # Looping through all the versions of each part, starting with the
            # oldest version.  Basically, it gets the most recent 
            # set of grade data for each part.
            my @submissions = ();
	    for(my $Version=1; $Version<=$LatestVersion; $Version++) {
                foreach my $part (split(/\:/,$data->{$sequence.':'.
                                                     $problemID.
                                                     ':parts'})) {

                    if(!defined($input->{"$name:$Version:$problem".
                                         ":resource.$part.solved"})) {
                        # No grade for this submission, so skip
                        next;
                    }

                    my $tries=0;
                    my $code=' ';
                    my $awarded=0;

                    $tries = $input->{$name.':'.$Version.':'.$problem.
                                      ':resource.'.$part.'.tries'};
                    $awarded = $input->{$name.':'.$Version.':'.$problem.
                                        ':resource.'.$part.'.awarded'};

                    $partData{$part.':awarded'}=($awarded) ? $awarded : 0;
                    $partData{$part.':tries'}=($tries) ? $tries : 0;

                    $partData{$part.':timestamp'}=$input->{$name.':'.$Version.':'.
                                                           $problem.
                                                           ':timestamp'};
                    if(!$input->{$name.':'.$Version.':'.$problem.':resource.'.$part.
                                 '.previous'}) {
                        foreach my $response (split(':',
                                                   $data->{$sequence.':'.
                                                           $problemID.':'.
                                                           $part.':responseIDs'})) {
                            @submissions=($input->{$name.':'.$Version.':'.
                                                   $problem.
                                                   ':resource.'.$part.'.'.
                                                   $response.'.submission'},
                                          @submissions);
                        }
                    }

                    my $val = $input->{$name.':'.$Version.':'.$problem.
                                       ':resource.'.$part.'.solved'};
                    if    ($val eq 'correct_by_student')   {$code = '*';} 
                    elsif ($val eq 'correct_by_override')  {$code = '+';}
                    elsif ($val eq 'incorrect_attempted')  {$code = '.';} 
                    elsif ($val eq 'incorrect_by_override'){$code = '-';}
                    elsif ($val eq 'excused')              {$code = 'x';}
                    elsif ($val eq 'ungraded_attempted')   {$code = '#';}
                    else                                   {$code = ' ';}
                    $partData{$part.':code'}=$code;
                }
            }

            foreach my $part (split(/\:/,$data->{$sequence.':'.$problemID.
                                                 ':parts'})) {
                $output->{$name.':'.$problemID.':'.$part.':wrong'} = 
                    $partData{$part.':tries'};
		$allkeys{$name.':'.$problemID.':'.$part.':wrong'}++;

                if($partData{$part.':code'} eq '*') {
                    $output->{$name.':'.$problemID.':'.$part.':wrong'}--;
                    $problemsCorrect++;
                } elsif($partData{$part.':code'} eq '+') {
                    $output->{$name.':'.$problemID.':'.$part.':wrong'}--;
                    $problemsCorrect++;
                }

                $output->{$name.':'.$problemID.':'.$part.':tries'} = 
                    $partData{$part.':tries'};
                $output->{$name.':'.$problemID.':'.$part.':code'} =
                    $partData{$part.':code'};
                $output->{$name.':'.$problemID.':'.$part.':awarded'} =
                    $partData{$part.':awarded'};
		$allkeys{$name.':'.$problemID.':'.$part.':tries'}++;
		$allkeys{$name.':'.$problemID.':'.$part.':code'}++;
		$allkeys{$name.':'.$problemID.':'.$part.':awarded'}++;

                $totalAwarded += $partData{$part.':awarded'};
                $output->{$name.':'.$problemID.':'.$part.':timestamp'} =
                    $partData{$part.':timestamp'};
		$allkeys{$name.':'.$problemID.':'.$part.':timestamp'}++;

                foreach my $response (split(':', $data->{$sequence.':'.
                                                         $problemID.':'.
                                                         $part.':responseIDs'})) {
                    $output->{$name.':'.$problemID.':'.$part.':'.$response.
                              ':submission'}=join(':::',@submissions);
		    $allkeys{$name.':'.$problemID.':'.$part.':'.$response.
			     ':submission'}++;
                }

                if($partData{$part.':code'} ne 'x') {
                    $totalProblems++;
                }
            }
        }

        $output->{$name.':'.$sequence.':problemsCorrect'} = $problemsCorrect;
	$allkeys{$name.':'.$sequence.':problemsCorrect'}++;
        $problemsSolved += $problemsCorrect;
	$problemsCorrect=0;
    }

    $output->{$name.':problemsSolved'} = $problemsSolved;
    $output->{$name.':totalProblems'} = $totalProblems;
    $output->{$name.':totalAwarded'} = $totalAwarded;
    $allkeys{$name.':problemsSolved'}++;
    $allkeys{$name.':totalProblems'}++;
    $allkeys{$name.':totalAwarded'}++;

    $output->{$name.':keys'} = join(':::', keys(%allkeys));

    return;
}

sub LoadDiscussion {
    my ($courseID)=@_;
    my %Discuss=();
    my %contrib=&Apache::lonnet::dump(
                $courseID,
                $ENV{'course.'.$courseID.'.domain'},
                $ENV{'course.'.$courseID.'.num'});
				 
    #my %contrib=&DownloadCourseInformation($name, $courseID, 0);

    foreach my $temp(keys %contrib) {
	if ($temp=~/^version/) {
	    my $ver=$contrib{$temp};
	    my ($dummy,$prb)=split(':',$temp);
	    for (my $idx=1; $idx<=$ver; $idx++ ) {
		my $name=$contrib{"$idx:$prb:sendername"};
		$Discuss{"$name:$prb"}=$idx;	
	    }
	}
    }       

    return \%Discuss;
}

# ----- END PROCESSING FUNCTIONS ---------------------------------------

=pod

=head1 HELPER FUNCTIONS

These are just a couple of functions do various odd and end 
jobs.  There was also a couple of bulk functions added.  These are
&DownloadStudentCourseData(), &DownloadStudentCourseDataSeparate(), and
&CheckForResidualDownload().  These functions now act as the interface
for downloading student course data.  The statistical modules should
no longer make the calls to dump and download and process etc.  They
make calls to these bulk functions to get their data.

=cut

# ----- HELPER FUNCTIONS -----------------------------------------------

sub CheckDateStampError {
    my ($courseData, $cache, $name)=@_;
    if($courseData->{$name.':UpToDate'} eq 'true') {
        $cache->{$name.':lastDownloadTime'} = 
            $courseData->{$name.':lastDownloadTime'};
        if($courseData->{$name.':lastDownloadTime'} eq 'Not downloaded') {
            $cache->{$name.':updateTime'} = ' Not updated';
        } else {
            $cache->{$name.':updateTime'}=
                localtime($courseData->{$name.':lastDownloadTime'});
        }
        return 0;
    }

    $cache->{$name.':lastDownloadTime'}=$courseData->{$name.':lastDownloadTime'};
    if($courseData->{$name.':lastDownloadTime'} eq 'Not downloaded') {
        $cache->{$name.':updateTime'} = ' Not updated';
    } else {
        $cache->{$name.':updateTime'}=
            localtime($courseData->{$name.':lastDownloadTime'});
    }

    if(defined($courseData->{$name.':error'})) {
        $cache->{$name.':error'}=$courseData->{$name.':error'};
        return 0;
    }

    return 1;
}

=pod

=item &ProcessFullName()

Takes lastname, generation, firstname, and middlename (or some partial
set of this data) and returns the full name version as a string.  Format
is Lastname generation, firstname middlename or a subset of this.

=cut

sub ProcessFullName {
    my ($lastname, $generation, $firstname, $middlename)=@_;
    my $Str = '';

    # Strip whitespace preceeding & following name components.
    $lastname   =~ s/(\s+$|^\s+)//g;
    $generation =~ s/(\s+$|^\s+)//g;
    $firstname  =~ s/(\s+$|^\s+)//g;
    $middlename =~ s/(\s+$|^\s+)//g;

    if($lastname ne '') {
	$Str .= $lastname;
	$Str .= ' '.$generation if ($generation ne '');
	$Str .= ',';
        $Str .= ' '.$firstname  if ($firstname ne '');
        $Str .= ' '.$middlename if ($middlename ne '');
    } else {
        $Str .= $firstname      if ($firstname ne '');
        $Str .= ' '.$middlename if ($middlename ne '');
        $Str .= ' '.$generation if ($generation ne '');
    }

    return $Str;
}

=pod

=item &TestCacheData()

Determine if the cache database can be accessed with a tie.  It waits up to
ten seconds before returning failure.  This function exists to help with
the problems with stopping the data download.  When an abort occurs and the
user quickly presses a form button and httpd child is created.  This
child needs to wait for the other to finish (hopefully within ten seconds).

=over 4

Input: $ChartDB

$ChartDB: The name of the cache database to be opened

Output: -1, 0, 1

-1: Could not tie database
 0: Use cached data
 1: New cache database created, use that.

=back

=cut

sub TestCacheData {
    my ($ChartDB,$isRecalculate,$totalDelay)=@_;
    my $isCached=-1;
    my %testData;
    my $tieTries=0;

    if(!defined($totalDelay)) {
        $totalDelay = 10;
    }

    if ((-e "$ChartDB") && (!$isRecalculate)) {
	$isCached = 1;
    } else {
	$isCached = 0;
    }

    while($tieTries < $totalDelay) {
        my $result=0;
        if($isCached) {
            $result=tie(%testData,'GDBM_File',$ChartDB,&GDBM_READER(),0640);
        } else {
            $result=tie(%testData,'GDBM_File',$ChartDB,&GDBM_NEWDB(),0640);
        }
        if($result) {
            last;
        }
        $tieTries++;
        sleep 1;
    }
    if($tieTries >= $totalDelay) {
        return -1;
    }

    untie(%testData);

    return $isCached;
}

sub DownloadStudentCourseData {
    my ($students,$checkDate,$cacheDB,$extract,$status,$courseID,$r,$c)=@_;

    my $title = 'LON-CAPA Statistics';
    my $heading = 'Download and Process Course Data';
    my $studentCount = scalar(@$students);

    my $WhatIWant;
    $WhatIWant = '(^version:|';
    $WhatIWant .= '^\d+:.+?:(resource\.\d+\.';
    $WhatIWant .= '(solved|tries|previous|awarded|(\d+\.submission))\s*$';
    $WhatIWant .= '|timestamp)';
    $WhatIWant .= ')';
#    $WhatIWant = '.';

    if($status eq 'true') {
        &Apache::lonhtmlcommon::Create_PrgWin($r, $title, $heading);
    }

    my $displayString;
    my $count=0;
    foreach (@$students) {
        my %cache;

        if($c->aborted()) { return 'Aborted'; }

        if($status eq 'true') {
            $count++;
            my $displayString = $count.'/'.$studentCount.': '.$_;
            &Apache::lonhtmlcommon::Update_PrgWin($displayString, $r);
        }

        my $downloadTime='Not downloaded';
        my $needUpdate = 'false';
        if($checkDate eq 'true'  && 
           tie(%cache,'GDBM_File',$cacheDB,&GDBM_READER(),0640)) {
            $downloadTime = $cache{$_.':lastDownloadTime'};
            $needUpdate = $cache{'ResourceUpdated'};
            untie(%cache);
        }

        if($c->aborted()) { return 'Aborted'; }

        if($needUpdate eq 'true') {
            $downloadTime = 'Not downloaded';
	}
	my $courseData = 
	    &DownloadCourseInformation($_, $courseID, $downloadTime, 
				       $WhatIWant);
	if(tie(%cache,'GDBM_File',$cacheDB,&GDBM_WRCREAT(),0640)) {
	    foreach my $key (keys(%$courseData)) {
		if($key =~ /^(con_lost|error|no_such_host)/i) {
		    $courseData->{$_.':error'} = 'No course data for '.$_;
		    last;
		}
	    }
	    if($extract eq 'true') {
		&ExtractStudentData($courseData, \%cache, \%cache, $_);
	    } else {
		&ProcessStudentData(\%cache, $courseData, $_);
	    }
	    untie(%cache);
	} else {
	    next;
	}
    }
    if($status eq 'true') { &Apache::lonhtmlcommon::Close_PrgWin($r); }

    return 'OK';
}

sub DownloadStudentCourseDataSeparate {
    my ($students,$checkDate,$cacheDB,$extract,$status,$courseID,$r,$c)=@_;
    my $residualFile = '/home/httpd/perl/tmp/'.$courseID.'DownloadFile.db';
    my $title = 'LON-CAPA Statistics';
    my $heading = 'Download Course Data';

    my $WhatIWant;
    $WhatIWant = '(^version:|';
    $WhatIWant .= '^\d+:.+?:(resource\.\d+\.';
    $WhatIWant .= '(solved|tries|previous|awarded|(\d+\.submission))\s*$';
    $WhatIWant .= '|timestamp)';
    $WhatIWant .= ')';

    &CheckForResidualDownload($cacheDB, 'true', 'true', $courseID, $r, $c);

    my $studentCount = scalar(@$students);
    if($status eq 'true') {
        &Apache::lonhtmlcommon::Create_PrgWin($r, $title, $heading);
    }
    my $count=0;
    my $displayString='';
    foreach (@$students) {
        if($c->aborted()) {
            return 'Aborted';
        }

        if($status eq 'true') {
            $count++;
            $displayString = $count.'/'.$studentCount.': '.$_;
            &Apache::lonhtmlcommon::Update_PrgWin($displayString, $r);
        }

        my %cache;
        my $downloadTime='Not downloaded';
        my $needUpdate = 'false';
        if($checkDate eq 'true'  && 
           tie(%cache,'GDBM_File',$cacheDB,&GDBM_READER(),0640)) {
            $downloadTime = $cache{$_.':lastDownloadTime'};
            $needUpdate = $cache{'ResourceUpdated'};
            untie(%cache);
        }

        if($c->aborted()) {
            return 'Aborted';
        }

        if($needUpdate eq 'true') {
            $downloadTime = 'Not downloaded';
	}

        my $error = 0;
        my $courseData = 
            &DownloadCourseInformation($_, $courseID, $downloadTime,
                                       $WhatIWant);
        my %downloadData;
        unless(tie(%downloadData,'GDBM_File',$residualFile,
                   &GDBM_WRCREAT(),0640)) {
            return 'Failed to tie temporary download hash.';
        }
        foreach my $key (keys(%$courseData)) {
            $downloadData{$key} = $courseData->{$key};
            if($key =~ /^(con_lost|error|no_such_host)/i) {
                $error = 1;
                last;
            }
        }
        if($error) {
            foreach my $deleteKey (keys(%$courseData)) {
                delete $downloadData{$deleteKey};
            }
            $downloadData{$_.':error'} = 'No course data for '.$_;
        }
        untie(%downloadData);
    }
    if($status eq 'true') { &Apache::lonhtmlcommon::Close_PrgWin($r); }

    return &CheckForResidualDownload($cacheDB, 'true', 'true', 
                                     $courseID, $r, $c);
}

sub CheckForResidualDownload {
    my ($cacheDB,$extract,$status,$courseID,$r,$c)=@_;

    my $residualFile = '/home/httpd/perl/tmp/'.$courseID.'DownloadFile.db';
    if(!-e $residualFile) {
        return 'OK';
    }

    my %downloadData;
    my %cache;
    unless(tie(%downloadData,'GDBM_File',$residualFile,&GDBM_READER(),0640)) {
        return 'Can not tie database for check for residual download: tempDB';
    }
    unless(tie(%cache,'GDBM_File',$cacheDB,&GDBM_WRCREAT(),0640)) {
        untie(%downloadData);
        return 'Can not tie database for check for residual download: cacheDB';
    }

    my @students=();
    my %checkStudent;
    my $key;
    while(($key, undef) = each %downloadData) {
        my @temp = split(':', $key);
        my $student = $temp[0].':'.$temp[1];
        if(!defined($checkStudent{$student})) {
            $checkStudent{$student}++;
            push(@students, $student);
        }
    }

    my $heading = 'Process Course Data';
    my $title = 'LON-CAPA Statistics';
    my $studentCount = scalar(@students);
    if($status eq 'true') {
        &Apache::lonhtmlcommon::Create_PrgWin($r, $title, $heading);
    }

    my $count=1;
    foreach my $name (@students) {
        last if($c->aborted());

        if($status eq 'true') {
            my $displayString = $count.'/'.$studentCount.': '.$name;
            &Apache::lonhtmlcommon::Update_PrgWin($displayString, $r);
        }

        if($extract eq 'true') {
            &ExtractStudentData(\%downloadData, \%cache, \%cache, $name);
        } else {
            &ProcessStudentData(\%cache, \%downloadData, $name);
        }
        $count++;
    }

    if($status eq 'true') { &Apache::lonhtmlcommon::Close_PrgWin($r); }

    untie(%cache);
    untie(%downloadData);

    if(!$c->aborted()) {
        my @files = ($residualFile);
        unlink(@files);
    }

    return 'OK';
}

################################################
################################################

=pod

=item &get_classlist();

Retrieve the classist of a given class or of the current class.  Student
information is returned from the classlist.db file and, if needed,
from the students environment.

Optional arguments are $cid, $cdom, and $cnum (course id, course domain,
and course number, respectively).  Any omitted arguments will be taken 
from the current environment ($ENV{'request.course.id'},
$ENV{'course.'.$cid.'.domain'}, and $ENV{'course.'.$cid.'.num'}).

Returns a reference to a hash which contains:
 keys    '$sname:$sdom'
 values  [$end,$start,$id,$section,$fullname]

=cut

################################################
################################################

sub get_classlist {
    my ($cid,$cdom,$cnum) = @_;
    $cid = $cid || $ENV{'request.course.id'};
    $cdom = $cdom || $ENV{'course.'.$cid.'.domain'};
    $cnum = $cnum || $ENV{'course.'.$cid.'.num'};
   my $now = time;
    #
    my %classlist=&Apache::lonnet::dump('classlist',$cdom,$cnum);
    while (my ($student,$info) = each(%classlist)) {
        return undef if ($student =~ /^(con_lost|error|no_such_host)/i);
        my ($sname,$sdom) = split(/:/,$student);
        my @Values = split(/:/,$info);
        my ($end,$start,$id,$section,$fullname);
        if (@Values > 2) {
            ($end,$start,$id,$section,$fullname) = @Values;
        } else { # We have to get the data ourselves
            ($end,$start) = @Values;
            $section = &Apache::lonnet::getsection($sdom,$sname,$cid);
            my %info=&Apache::lonnet::get('environment',
                                          ['firstname','middlename',
                                           'lastname','generation','id'],
                                          $sdom, $sname);
            my ($tmp) = keys(%info);
            if ($tmp =~/^(con_lost|error|no_such_host)/i) {
                $fullname = 'not available';
                $id = 'not available';
                &Apache::lonnet::logthis('unable to retrieve environment '.
                                         'for '.$sname.':'.$sdom);
            } else {
                $fullname = &ProcessFullName(@info{qw/lastname generation 
                                                       firstname middlename/});
                $id = $info{'id'};
            }
            # Update the classlist with this students information
            if ($fullname ne 'not available') {
                my $enrolldata = join(':',$end,$start,$id,$section,$fullname);
                my $reply=&Apache::lonnet::cput('classlist',
                                                {$student => $enrolldata},
                                                $cdom,$cnum);
                if ($reply !~ /^(ok|delayed)/) {
                    &Apache::lonnet::logthis('Unable to update classlist for '.
                                             'student '.$sname.':'.$sdom.
                                             ' error:'.$reply);
                }
            }
        }
        my $status='Expired';
        if(((!$end) || $now < $end) && ((!$start) || ($now > $start))) {
            $status='Active';
        }
        $classlist{$student} = 
            [$sdom,$sname,$end,$start,$id,$section,$fullname,$status];
    }
    if (wantarray()) {
        return (\%classlist,['domain','username','end','start','id',
                             'section','fullname','status']);
    } else {
        return \%classlist;
    }
}

# ----- END HELPER FUNCTIONS --------------------------------------------

1;
__END__


